@@ -35,10 +35,10 @@ from samba.netcmd import (
     Command,
     CommandError,
     Option,
-    )
+)
+
+RE_RANGED_RESULT = re.compile("^([^;]+);range=(\d+)-(\d+|\*)$")
 
-global summary
-summary = {}
 
 class LDAPBase(object):
 
@@ -48,7 +48,7 @@ class LDAPBase(object):
                  outf=sys.stdout, errf=sys.stderr, skip_missing_dn=True):
         ldb_options = []
         samdb_url = host
-        if not "://" in host:
+        if "://" not in host:
             if os.path.isfile(host):
                 samdb_url = "tdb://%s" % host
             else:
@@ -86,43 +86,41 @@ class LDAPBase(object):
         # when compare content of two DCs. Uncomment for DEBUG purposes.
         if self.two_domains and not self.quiet:
             self.outf.write("\n* Place-holders for %s:\n" % self.host)
-            self.outf.write(4*" " + "${DOMAIN_DN}      => %s\n" %
-                self.base_dn)
-            self.outf.write(4*" " + "${DOMAIN_NETBIOS} => %s\n" %
-                self.domain_netbios)
-            self.outf.write(4*" " + "${SERVER_NAME}     => %s\n" %
-                self.server_names)
-            self.outf.write(4*" " + "${DOMAIN_NAME}    => %s\n" %
-                self.domain_name)
+            self.outf.write(4 * " " + "${DOMAIN_DN}      => %s\n" %
+                            self.base_dn)
+            self.outf.write(4 * " " + "${DOMAIN_NETBIOS} => %s\n" %
+                            self.domain_netbios)
+            self.outf.write(4 * " " + "${SERVER_NAME}     => %s\n" %
+                            self.server_names)
+            self.outf.write(4 * " " + "${DOMAIN_NAME}    => %s\n" %
+                            self.domain_name)
 
     def find_domain_sid(self):
         res = self.ldb.search(base=self.base_dn, expression="(objectClass=*)", scope=SCOPE_BASE)
-        return ndr_unpack(security.dom_sid,res[0]["objectSid"][0])
+        return ndr_unpack(security.dom_sid, res[0]["objectSid"][0])
 
     def find_servers(self):
         """
         """
         res = self.ldb.search(base="OU=Domain Controllers,%s" % self.base_dn,
-                scope=SCOPE_SUBTREE, expression="(objectClass=computer)", attrs=["cn"])
+                              scope=SCOPE_SUBTREE, expression="(objectClass=computer)", attrs=["cn"])
         assert len(res) > 0
-        srv = []
-        for x in res:
-            srv.append(x["cn"][0])
-        return srv
+        return [str(x["cn"][0]) for x in res]
 
     def find_netbios(self):
         res = self.ldb.search(base="CN=Partitions,%s" % self.config_dn,
-                scope=SCOPE_SUBTREE, attrs=["nETBIOSName"])
+                              scope=SCOPE_SUBTREE, attrs=["nETBIOSName"])
         assert len(res) > 0
         for x in res:
-            if "nETBIOSName" in x.keys():
+            if "nETBIOSName" in x:
                 return x["nETBIOSName"][0]
 
     def object_exists(self, object_dn):
         res = None
         try:
             res = self.ldb.search(base=object_dn, scope=SCOPE_BASE)
-        except LdbError, (enum, estr):
+        except LdbError as e2:
+            (enum, estr) = e2.args
             if enum == ERR_NO_SUCH_OBJECT:
                 return False
             raise
@@ -131,7 +129,7 @@ class LDAPBase(object):
     def delete_force(self, object_dn):
         try:
             self.ldb.delete(object_dn)
-        except Ldb.LdbError, e:
+        except Ldb.LdbError as e:
             assert "No such object" in str(e)
 
     def get_attribute_name(self, key):
@@ -139,9 +137,7 @@ class LDAPBase(object):
             It resolved ranged results e.g. member;range=0-1499
         """
 
-        r = re.compile("^([^;]+);range=(\d+)-(\d+|\*)$")
-
-        m = r.match(key)
+        m = RE_RANGED_RESULT.match(key)
         if m is None:
             return key
 
@@ -152,9 +148,7 @@ class LDAPBase(object):
             It resolved ranged results e.g. member;range=0-1499
         """
 
-        r = re.compile("^([^;]+);range=(\d+)-(\d+|\*)$")
-
-        m = r.match(key)
+        m = RE_RANGED_RESULT.match(key)
         if m is None:
             # no range, just return the values
             return vals
@@ -175,8 +169,8 @@ class LDAPBase(object):
             fm = None
             fvals = None
 
-            for key in res.keys():
-                m = r.match(key)
+            for key in res:
+                m = RE_RANGED_RESULT.match(key)
 
                 if m is None:
                     continue
@@ -209,13 +203,15 @@ class LDAPBase(object):
         res = dict(res[0])
         # 'Dn' element is not iterable and we have it as 'distinguishedName'
         del res["dn"]
-        for key in res.keys():
-            vals = list(res[key])
-            del res[key]
-            name = self.get_attribute_name(key)
-            res[name] = self.get_attribute_values(object_dn, key, vals)
 
-        return res
+        attributes = {}
+        for key, vals in res.items():
+            name = self.get_attribute_name(key)
+            # sort vals and return a list, help to compare
+            vals = sorted(vals)
+            attributes[name] = self.get_attribute_values(object_dn, key, vals)
+
+        return attributes
 
     def get_descriptor_sddl(self, object_dn):
         res = self.ldb.search(base=object_dn, scope=SCOPE_BASE, attrs=["nTSecurityDescriptor"])
@@ -257,10 +253,11 @@ class LDAPBase(object):
                               expression="(objectSid=*)", scope=SCOPE_SUBTREE, attrs=["objectSid", "sAMAccountName"])
         for item in res:
             try:
-                self.sid_map["%s" % ndr_unpack(security.dom_sid, item["objectSid"][0])] = item["sAMAccountName"][0]
+                self.sid_map["%s" % ndr_unpack(security.dom_sid, item["objectSid"][0])] = str(item["sAMAccountName"][0])
             except KeyError:
                 pass
 
+
 class Descriptor(object):
     def __init__(self, connection, dn, outf=sys.stdout, errf=sys.stderr):
         self.outf = outf
@@ -302,9 +299,9 @@ class Descriptor(object):
     def diff_1(self, other):
         res = ""
         if len(self.dacl_list) != len(other.dacl_list):
-            res += 4*" " + "Difference in ACE count:\n"
-            res += 8*" " + "=> %s\n" % len(self.dacl_list)
-            res += 8*" " + "=> %s\n" % len(other.dacl_list)
+            res += 4 * " " + "Difference in ACE count:\n"
+            res += 8 * " " + "=> %s\n" % len(self.dacl_list)
+            res += 8 * " " + "=> %s\n" % len(other.dacl_list)
         #
         i = 0
         flag = True
@@ -325,27 +322,25 @@ class Descriptor(object):
             self_ace_fixed = "%s" % self.fix_sid(self_ace)
             other_ace_fixed = "%s" % other.fix_sid(other_ace)
             if self_ace_fixed != other_ace_fixed:
-                res += "%60s * %s\n" % ( self_ace_fixed, other_ace_fixed )
+                res += "%60s * %s\n" % (self_ace_fixed, other_ace_fixed)
                 flag = False
             else:
-                res += "%60s | %s\n" % ( self_ace_fixed, other_ace_fixed )
+                res += "%60s | %s\n" % (self_ace_fixed, other_ace_fixed)
             i += 1
         return (flag, res)
 
     def diff_2(self, other):
         res = ""
         if len(self.dacl_list) != len(other.dacl_list):
-            res += 4*" " + "Difference in ACE count:\n"
-            res += 8*" " + "=> %s\n" % len(self.dacl_list)
-            res += 8*" " + "=> %s\n" % len(other.dacl_list)
+            res += 4 * " " + "Difference in ACE count:\n"
+            res += 8 * " " + "=> %s\n" % len(self.dacl_list)
+            res += 8 * " " + "=> %s\n" % len(other.dacl_list)
         #
         common_aces = []
         self_aces = []
         other_aces = []
-        self_dacl_list_fixed = []
-        other_dacl_list_fixed = []
-        [self_dacl_list_fixed.append( self.fix_sid(ace) ) for ace in self.dacl_list]
-        [other_dacl_list_fixed.append( other.fix_sid(ace) ) for ace in other.dacl_list]
+        self_dacl_list_fixed = [self.fix_sid(ace) for ace in self.dacl_list]
+        other_dacl_list_fixed = [other.fix_sid(ace) for ace in other.dacl_list]
         for ace in self_dacl_list_fixed:
             try:
                 other_dacl_list_fixed.index(ace)
@@ -355,9 +350,9 @@ class Descriptor(object):
                 common_aces.append(ace)
         self_aces = sorted(self_aces)
         if len(self_aces) > 0:
-            res += 4*" " + "ACEs found only in %s:\n" % self.con.host
+            res += 4 * " " + "ACEs found only in %s:\n" % self.con.host
             for ace in self_aces:
-                res += 8*" " + ace + "\n"
+                res += 8 * " " + ace + "\n"
         #
         for ace in other_dacl_list_fixed:
             try:
@@ -368,17 +363,18 @@ class Descriptor(object):
                 common_aces.append(ace)
         other_aces = sorted(other_aces)
         if len(other_aces) > 0:
-            res += 4*" " + "ACEs found only in %s:\n" % other.con.host
+            res += 4 * " " + "ACEs found only in %s:\n" % other.con.host
             for ace in other_aces:
-                res += 8*" " + ace + "\n"
+                res += 8 * " " + ace + "\n"
         #
         common_aces = sorted(list(set(common_aces)))
         if self.con.verbose:
-            res += 4*" " + "ACEs found in both:\n"
+            res += 4 * " " + "ACEs found in both:\n"
             for ace in common_aces:
-                res += 8*" " + ace + "\n"
+                res += 8 * " " + ace + "\n"
         return (self_aces == [] and other_aces == [], res)
 
+
 class LDAPObject(object):
     def __init__(self, connection, dn, summary, filter_list,
                  outf=sys.stdout, errf=sys.stderr):
@@ -438,7 +434,7 @@ class LDAPObject(object):
                 "uSNChanged",
                 "uSNCreated",
                 "uSNLastObjRem",
-                # "whenChanged", # This is implicitly replicated
+                "whenChanged",  # This is implicitly replicated, but may diverge on updates of non-replicated attributes
         ]
         self.ignore_attributes = self.non_replicated_attributes
         self.ignore_attributes += ["msExchServer1HighestUSN"]
@@ -453,13 +449,18 @@ class LDAPObject(object):
         # Two domains - two domain controllers
 
         if self.two_domains:
-            self.ignore_attributes +=  [
-                "objectCategory", "objectGUID", "objectSid", "whenCreated", "whenChanged", "pwdLastSet", "uSNCreated", "creationTime",
-                "modifiedCount", "priorSetTime", "rIDManagerReference", "gPLink", "ipsecNFAReference",
-                "fRSPrimaryMember", "fSMORoleOwner", "masteredBy", "ipsecOwnersReference", "wellKnownObjects",
-                "badPwdCount", "ipsecISAKMPReference", "ipsecFilterReference", "msDs-masteredBy", "lastSetTime",
-                "ipsecNegotiationPolicyReference", "subRefs", "gPCFileSysPath", "accountExpires", "invocationId",
-                "operatingSystemVersion", "oEMInformation",
+            self.ignore_attributes += [
+                "objectCategory", "objectGUID", "objectSid", "whenCreated",
+                "whenChanged", "pwdLastSet", "uSNCreated", "creationTime",
+                "modifiedCount", "priorSetTime", "rIDManagerReference",
+                "gPLink", "ipsecNFAReference", "fRSPrimaryMember",
+                "fSMORoleOwner", "masteredBy", "ipsecOwnersReference",
+                "wellKnownObjects", "otherWellKnownObjects", "badPwdCount",
+                "ipsecISAKMPReference", "ipsecFilterReference",
+                "msDs-masteredBy", "lastSetTime",
+                "ipsecNegotiationPolicyReference", "subRefs", "gPCFileSysPath",
+                "accountExpires", "invocationId", "operatingSystemVersion",
+                "oEMInformation",
                 # After Exchange preps
                 "targetAddress", "msExchMailboxGuid", "siteFolderGUID"]
             #
@@ -473,42 +474,49 @@ class LDAPObject(object):
                 "msExchHomeRoutingGroup", "msExchResponsibleMTAServer", "siteFolderServer", "msExchRoutingMasterDN",
                 "msExchRoutingGroupMembersBL", "homeMDBBL", "msExchHomePublicMDB", "msExchOwningServer", "templateRoots",
                 "addressBookRoots", "msExchPolicyRoots", "globalAddressList", "msExchOwningPFTree",
-                "msExchResponsibleMTAServerBL", "msExchOwningPFTreeBL",]
+                "msExchResponsibleMTAServerBL", "msExchOwningPFTreeBL",
+                # After 2012 R2 functional preparation
+                "msDS-MembersOfResourcePropertyListBL",
+                "msDS-ValueTypeReference",
+                "msDS-MembersOfResourcePropertyList",
+                "msDS-ValueTypeReferenceBL",
+                "msDS-ClaimTypeAppliesToClass",
+            ]
             self.dn_attributes = [x.upper() for x in self.dn_attributes]
             #
             # Attributes that contain the Domain name e.g. 'samba.org'
             self.domain_attributes = [
                 "proxyAddresses", "mail", "userPrincipalName", "msExchSmtpFullyQualifiedDomainName",
-                "dnsHostName", "networkAddress", "dnsRoot", "servicePrincipalName",]
+                "dnsHostName", "networkAddress", "dnsRoot", "servicePrincipalName", ]
             self.domain_attributes = [x.upper() for x in self.domain_attributes]
             #
             # May contain DOMAIN_NETBIOS and SERVER_NAME
-            self.servername_attributes = [ "distinguishedName", "name", "CN", "sAMAccountName", "dNSHostName",
-                "servicePrincipalName", "rIDSetReferences", "serverReference", "serverReferenceBL",
-                "msDS-IsDomainFor", "interSiteTopologyGenerator",]
+            self.servername_attributes = ["distinguishedName", "name", "CN", "sAMAccountName", "dNSHostName",
+                                          "servicePrincipalName", "rIDSetReferences", "serverReference", "serverReferenceBL",
+                                          "msDS-IsDomainFor", "interSiteTopologyGenerator", ]
             self.servername_attributes = [x.upper() for x in self.servername_attributes]
             #
-            self.netbios_attributes = [ "servicePrincipalName", "CN", "distinguishedName", "nETBIOSName", "name",]
+            self.netbios_attributes = ["servicePrincipalName", "CN", "distinguishedName", "nETBIOSName", "name", ]
             self.netbios_attributes = [x.upper() for x in self.netbios_attributes]
             #
-            self.other_attributes = [ "name", "DC",]
+            self.other_attributes = ["name", "DC", ]
             self.other_attributes = [x.upper() for x in self.other_attributes]
         #
-        self.ignore_attributes = [x.upper() for x in self.ignore_attributes]
+        self.ignore_attributes = set([x.upper() for x in self.ignore_attributes])
 
     def log(self, msg):
         """
         Log on the screen if there is no --quiet option set
         """
         if not self.quiet:
-            self.outf.write(msg+"\n")
+            self.outf.write(msg +"\n")
 
     def fix_dn(self, s):
         res = "%s" % s
         if not self.two_domains:
             return res
         if res.upper().endswith(self.con.base_dn.upper()):
-            res = res[:len(res)-len(self.con.base_dn)] + "${DOMAIN_DN}"
+            res = res[:len(res) - len(self.con.base_dn)] + "${DOMAIN_DN}"
         return res
 
     def fix_domain_name(self, s):
@@ -550,42 +558,33 @@ class LDAPObject(object):
         else:
             raise Exception("Unknown --view option value.")
         #
-        self.screen_output = res[1][:-1]
-        other.screen_output = res[1][:-1]
+        self.screen_output = res[1]
+        other.screen_output = res[1]
         #
         return res[0]
 
     def cmp_attrs(self, other):
         res = ""
-        self.unique_attrs = []
         self.df_value_attrs = []
-        other.unique_attrs = []
-        if self.attributes.keys() != other.attributes.keys():
-            #
-            title = 4*" " + "Attributes found only in %s:" % self.con.host
-            for x in self.attributes.keys():
-                if not x in other.attributes.keys() and \
-                not x.upper() in [q.upper() for q in other.ignore_attributes]:
-                    if title:
-                        res += title + "\n"
-                        title = None
-                    res += 8*" " + x + "\n"
-                    self.unique_attrs.append(x)
-            #
-            title = 4*" " + "Attributes found only in %s:" % other.con.host
-            for x in other.attributes.keys():
-                if not x in self.attributes.keys() and \
-                not x.upper() in [q.upper() for q in self.ignore_attributes]:
-                    if title:
-                        res += title + "\n"
-                        title = None
-                    res += 8*" " + x + "\n"
-                    other.unique_attrs.append(x)
-        #
-        missing_attrs = [x.upper() for x in self.unique_attrs]
-        missing_attrs += [x.upper() for x in other.unique_attrs]
-        title = 4*" " + "Difference in attribute values:"
-        for x in self.attributes.keys():
+
+        self_attrs = set([attr.upper() for attr in self.attributes])
+        other_attrs = set([attr.upper() for attr in other.attributes])
+
+        self_unique_attrs = self_attrs - other_attrs - other.ignore_attributes
+        if self_unique_attrs:
+            res += 4 * " " + "Attributes found only in %s:" % self.con.host
+            for x in self_unique_attrs:
+                res += 8 * " " + x + "\n"
+
+        other_unique_attrs = other_attrs - self_attrs - self.ignore_attributes
+        if other_unique_attrs:
+            res += 4 * " " + "Attributes found only in %s:" % other.con.host
+            for x in other_unique_attrs:
+                res += 8 * " " + x + "\n"
+
+        missing_attrs = self_unique_attrs & other_unique_attrs
+        title = 4 * " " + "Difference in attribute values:"
+        for x in self.attributes:
             if x.upper() in self.ignore_attributes or x.upper() in missing_attrs:
                 continue
             if isinstance(self.attributes[x], list) and isinstance(other.attributes[x], list):
@@ -654,25 +653,25 @@ class LDAPObject(object):
                     res += title + "\n"
                     title = None
                 if p and q:
-                    res += 8*" " + x + " => \n%s\n%s" % (p, q) + "\n"
+                    res += 8 * " " + x + " => \n%s\n%s" % (p, q) + "\n"
                 else:
-                    res += 8*" " + x + " => \n%s\n%s" % (self.attributes[x], other.attributes[x]) + "\n"
+                    res += 8 * " " + x + " => \n%s\n%s" % (self.attributes[x], other.attributes[x]) + "\n"
                 self.df_value_attrs.append(x)
         #
-        if self.unique_attrs + other.unique_attrs != []:
-            assert self.unique_attrs != other.unique_attrs
-        self.summary["unique_attrs"] += self.unique_attrs
+        if missing_attrs:
+            assert self_unique_attrs != other_unique_attrs
+        self.summary["unique_attrs"] += list(self_unique_attrs)
         self.summary["df_value_attrs"] += self.df_value_attrs
-        other.summary["unique_attrs"] += other.unique_attrs
-        other.summary["df_value_attrs"] += self.df_value_attrs # they are the same
+        other.summary["unique_attrs"] += list(other_unique_attrs)
+        other.summary["df_value_attrs"] += self.df_value_attrs  # they are the same
         #
-        self.screen_output = res[:-1]
-        other.screen_output = res[:-1]
+        self.screen_output = res
+        other.screen_output = res
         #
         return res == ""
 
 
-class LDAPBundel(object):
+class LDAPBundle(object):
 
     def __init__(self, connection, context, dn_list=None, filter_list=None,
                  outf=sys.stdout, errf=sys.stderr):
@@ -697,12 +696,12 @@ class LDAPBundel(object):
             self.context = context.upper()
             self.dn_list = self.get_dn_list(context)
         else:
-            raise Exception("Unknown initialization data for LDAPBundel().")
+            raise Exception("Unknown initialization data for LDAPBundle().")
         counter = 0
         while counter < len(self.dn_list) and self.two_domains:
             # Use alias reference
             tmp = self.dn_list[counter]
-            tmp = tmp[:len(tmp)-len(self.con.base_dn)] + "${DOMAIN_DN}"
+            tmp = tmp[:len(tmp) - len(self.con.base_dn)] + "${DOMAIN_DN}"
             tmp = tmp.replace("CN=%s" % self.con.domain_netbios, "CN=${DOMAIN_NETBIOS}")
             if len(self.con.server_names) == 1:
                 for x in self.con.server_names:
@@ -718,96 +717,83 @@ class LDAPBundel(object):
         Log on the screen if there is no --quiet option set
         """
         if not self.quiet:
-            self.outf.write(msg+"\n")
+            self.outf.write(msg + "\n")
 
     def update_size(self):
         self.size = len(self.dn_list)
         self.dn_list = sorted(self.dn_list)
 
-    def __eq__(self, other):
+    def diff(self, other):
         res = True
         if self.size != other.size:
-            self.log( "\n* DN lists have different size: %s != %s" % (self.size, other.size) )
+            self.log("\n* DN lists have different size: %s != %s" % (self.size, other.size))
             if not self.skip_missing_dn:
                 res = False
+
+        self_dns = set([q.upper() for q in self.dn_list])
+        other_dns = set([q.upper() for q in other.dn_list])
+
         #
         # This is the case where we want to explicitly compare two objects with different DNs.
         # It does not matter if they are in the same DC, in two DC in one domain or in two
         # different domains.
-        if self.search_scope != SCOPE_BASE:
-            title= "\n* DNs found only in %s:" % self.con.host
-            for x in self.dn_list:
-                if not x.upper() in [q.upper() for q in other.dn_list]:
-                    if title and not self.skip_missing_dn:
-                        self.log( title )
-                        title = None
-                        res = False
-                    self.log( 4*" " + x )
-                    self.dn_list[self.dn_list.index(x)] = ""
-            self.dn_list = [x for x in self.dn_list if x]
-            #
-            title= "\n* DNs found only in %s:" % other.con.host
-            for x in other.dn_list:
-                if not x.upper() in [q.upper() for q in self.dn_list]:
-                    if title and not self.skip_missing_dn:
-                        self.log( title )
-                        title = None
-                        res = False
-                    self.log( 4*" " + x )
-                    other.dn_list[other.dn_list.index(x)] = ""
-            other.dn_list = [x for x in other.dn_list if x]
-            #
-            self.update_size()
-            other.update_size()
-            assert self.size == other.size
-            assert sorted([x.upper() for x in self.dn_list]) == sorted([x.upper() for x in other.dn_list])
-        self.log( "\n* Objects to be compared: %s" % self.size )
+        if self.search_scope != SCOPE_BASE and not self.skip_missing_dn:
+
+            self_only = self_dns - other_dns  # missing in other
+            if self_only:
+                res = False
+                self.log("\n* DNs found only in %s:" % self.con.host)
+                for x in sorted(self_only):
+                    self.log(4 * " " + x)
+
+            other_only = other_dns - self_dns  # missing in self
+            if other_only:
+                res = False
+                self.log("\n* DNs found only in %s:" % other.con.host)
+                for x in sorted(other_only):
+                    self.log(4 * " " + x)
+
+        common_dns = self_dns & other_dns
+        self.log("\n* Objects to be compared: %d" % len(common_dns))
+
+        for dn in common_dns:
 
-        index = 0
-        while index < self.size:
-            skip = False
             try:
                 object1 = LDAPObject(connection=self.con,
-                                     dn=self.dn_list[index],
+                                     dn=dn,
                                      summary=self.summary,
                                      filter_list=self.filter_list,
                                      outf=self.outf, errf=self.errf)
-            except LdbError, (enum, estr):
-                if enum == ERR_NO_SUCH_OBJECT:
-                    self.log( "\n!!! Object not found: %s" % self.dn_list[index] )
-                    skip = True
-                raise
+            except LdbError as e:
+                self.log("LdbError for dn %s: %s" % (dn, e))
+                continue
+
             try:
                 object2 = LDAPObject(connection=other.con,
-                        dn=other.dn_list[index],
-                        summary=other.summary,
-                        filter_list=self.filter_list,
-                        outf=self.outf, errf=self.errf)
-            except LdbError, (enum, estr):
-                if enum == ERR_NO_SUCH_OBJECT:
-                    self.log( "\n!!! Object not found: %s" % other.dn_list[index] )
-                    skip = True
-                raise
-            if skip:
-                index += 1
+                                     dn=dn,
+                                     summary=other.summary,
+                                     filter_list=self.filter_list,
+                                     outf=self.outf, errf=self.errf)
+            except LdbError as e:
+                self.log("LdbError for dn %s: %s" % (dn, e))
                 continue
+
             if object1 == object2:
                 if self.con.verbose:
-                    self.log( "\nComparing:" )
-                    self.log( "'%s' [%s]" % (object1.dn, object1.con.host) )
-                    self.log( "'%s' [%s]" % (object2.dn, object2.con.host) )
-                    self.log( 4*" " + "OK" )
+                    self.log("\nComparing:")
+                    self.log("'%s' [%s]" % (object1.dn, object1.con.host))
+                    self.log("'%s' [%s]" % (object2.dn, object2.con.host))
+                    self.log(4 * " " + "OK")
             else:
-                self.log( "\nComparing:" )
-                self.log( "'%s' [%s]" % (object1.dn, object1.con.host) )
-                self.log( "'%s' [%s]" % (object2.dn, object2.con.host) )
-                self.log( object1.screen_output )
-                self.log( 4*" " + "FAILED" )
+                self.log("\nComparing:")
+                self.log("'%s' [%s]" % (object1.dn, object1.con.host))
+                self.log("'%s' [%s]" % (object2.dn, object2.con.host))
+                self.log(object1.screen_output)
+                self.log(4 * " " + "FAILED")
                 res = False
             self.summary = object1.summary
             other.summary = object2.summary
-            index += 1
-        #
+
         return res
 
     def get_dn_list(self, context):
@@ -836,17 +822,15 @@ class LDAPBundel(object):
         elif self.search_scope == "ONE":
             self.search_scope = SCOPE_ONELEVEL
         else:
-            raise StandardError("Wrong 'scope' given. Choose from: SUB, ONE, BASE")
+            raise ValueError("Wrong 'scope' given. Choose from: SUB, ONE, BASE")
         try:
             res = self.con.ldb.search(base=self.search_base, scope=self.search_scope, attrs=["dn"])
-        except LdbError, (enum, estr):
+        except LdbError as e3:
+            (enum, estr) = e3.args
             self.outf.write("Failed search of base=%s\n" % self.search_base)
             raise
         for x in res:
-           dn_list.append(x["dn"].get_linearized())
-        #
-        global summary
-        #
+            dn_list.append(x["dn"].get_linearized())
         return dn_list
 
     def print_summary(self):
@@ -854,12 +838,12 @@ class LDAPBundel(object):
         self.summary["df_value_attrs"] = list(set(self.summary["df_value_attrs"]))
         #
         if self.summary["unique_attrs"]:
-            self.log( "\nAttributes found only in %s:" % self.con.host )
-            self.log( "".join([str("\n" + 4*" " + x) for x in self.summary["unique_attrs"]]) )
+            self.log("\nAttributes found only in %s:" % self.con.host)
+            self.log("".join([str("\n" + 4 * " " + x) for x in self.summary["unique_attrs"]]))
         #
         if self.summary["df_value_attrs"]:
-            self.log( "\nAttributes with different values:" )
-            self.log( "".join([str("\n" + 4*" " + x) for x in self.summary["df_value_attrs"]]) )
+            self.log("\nAttributes with different values:")
+            self.log("".join([str("\n" + 4 * " " + x) for x in self.summary["df_value_attrs"]]))
             self.summary["df_value_attrs"] = []
 
 
@@ -873,38 +857,32 @@ class cmd_ldapcmp(Command):
         "credopts": options.CredentialsOptionsDouble,
     }
 
-    takes_optiongroups = {
-        "sambaopts": options.SambaOptions,
-        "versionopts": options.VersionOptions,
-        "credopts": options.CredentialsOptionsDouble,
-    }
-
     takes_args = ["URL1", "URL2", "context1?", "context2?", "context3?", "context4?", "context5?"]
 
     takes_options = [
         Option("-w", "--two", dest="two", action="store_true", default=False,
-            help="Hosts are in two different domains"),
+               help="Hosts are in two different domains"),
         Option("-q", "--quiet", dest="quiet", action="store_true", default=False,
-            help="Do not print anything but relay on just exit code"),
+               help="Do not print anything but relay on just exit code"),
         Option("-v", "--verbose", dest="verbose", action="store_true", default=False,
-            help="Print all DN pairs that have been compared"),
+               help="Print all DN pairs that have been compared"),
         Option("--sd", dest="descriptor", action="store_true", default=False,
-            help="Compare nTSecurityDescriptor attibutes only"),
+               help="Compare nTSecurityDescriptor attibutes only"),
         Option("--sort-aces", dest="sort_aces", action="store_true", default=False,
-            help="Sort ACEs before comparison of nTSecurityDescriptor attribute"),
-        Option("--view", dest="view", default="section",
-            help="Display mode for nTSecurityDescriptor results. Possible values: section or collision."),
+               help="Sort ACEs before comparison of nTSecurityDescriptor attribute"),
+        Option("--view", dest="view", default="section", choices=["section", "collision"],
+               help="Display mode for nTSecurityDescriptor results. Possible values: section or collision."),
         Option("--base", dest="base", default="",
-            help="Pass search base that will build DN list for the first DC."),
+               help="Pass search base that will build DN list for the first DC."),
         Option("--base2", dest="base2", default="",
-            help="Pass search base that will build DN list for the second DC. Used when --two or when compare two different DNs."),
-        Option("--scope", dest="scope", default="SUB",
-            help="Pass search scope that builds DN list. Options: SUB, ONE, BASE"),
+               help="Pass search base that will build DN list for the second DC. Used when --two or when compare two different DNs."),
+        Option("--scope", dest="scope", default="SUB", choices=["SUB", "ONE", "BASE"],
+               help="Pass search scope that builds DN list. Options: SUB, ONE, BASE"),
         Option("--filter", dest="filter", default="",
-            help="List of comma separated attributes to ignore in the comparision"),
+               help="List of comma separated attributes to ignore in the comparision"),
         Option("--skip-missing-dn", dest="skip_missing_dn", action="store_true", default=False,
-            help="Skip report and failure due to missing DNs in one server or another"),
-        ]
+               help="Skip report and failure due to missing DNs in one server or another"),
+    ]
 
     def run(self, URL1, URL2,
             context1=None, context2=None, context3=None, context4=None, context5=None,
@@ -951,21 +929,17 @@ class cmd_ldapcmp(Command):
             raise CommandError("You cannot set --verbose and --quiet together")
         if (not base and base2) or (base and not base2):
             raise CommandError("You need to specify both --base and --base2 at the same time")
-        if descriptor and view.upper() not in ["SECTION", "COLLISION"]:
-            raise CommandError("Invalid --view value. Choose from: section or collision")
-        if not scope.upper() in ["SUB", "ONE", "BASE"]:
-            raise CommandError("Invalid --scope value. Choose from: SUB, ONE, BASE")
 
         con1 = LDAPBase(URL1, creds, lp,
                         two=two, quiet=quiet, descriptor=descriptor, sort_aces=sort_aces,
-                        verbose=verbose,view=view, base=base, scope=scope,
-                        outf=self.outf, errf=self.errf)
+                        verbose=verbose, view=view, base=base, scope=scope,
+                        outf=self.outf, errf=self.errf, skip_missing_dn=skip_missing_dn)
         assert len(con1.base_dn) > 0
 
         con2 = LDAPBase(URL2, creds2, lp,
                         two=two, quiet=quiet, descriptor=descriptor, sort_aces=sort_aces,
                         verbose=verbose, view=view, base=base2, scope=scope,
-                        outf=self.outf, errf=self.errf)
+                        outf=self.outf, errf=self.errf, skip_missing_dn=skip_missing_dn)
         assert len(con2.base_dn) > 0
 
         filter_list = filter.split(",")
@@ -975,15 +949,15 @@ class cmd_ldapcmp(Command):
             if not quiet:
                 self.outf.write("\n* Comparing [%s] context...\n" % context)
 
-            b1 = LDAPBundel(con1, context=context, filter_list=filter_list,
+            b1 = LDAPBundle(con1, context=context, filter_list=filter_list,
                             outf=self.outf, errf=self.errf)
-            b2 = LDAPBundel(con2, context=context, filter_list=filter_list,
+            b2 = LDAPBundle(con2, context=context, filter_list=filter_list,
                             outf=self.outf, errf=self.errf)
 
-            if b1 == b2:
+            if b1.diff(b2):
                 if not quiet:
                     self.outf.write("\n* Result for [%s]: SUCCESS\n" %
-                        context)
+                                    context)
             else:
                 if not quiet:
                     self.outf.write("\n* Result for [%s]: FAILURE\n" % context)
